#!/usr/bin/python3
#
# Univention Management Console
#  Self Service Module for Users to generate their initial QR-Code Token
#
# Copyright 2020-2022 Univention GmbH
#
# https://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <https://www.gnu.org/licenses/>.

import requests
import univention.config_registry as ucr
from requests import ConnectionError
from six.moves import configparser
from univention.lib.i18n import Translation
from univention.management.console.base import Base
from univention.management.console.error import UMC_Error
from univention.management.console.modules.decorators import simple_response
from univention.management.console.log import MODULE

from .exceptions import ConfiguratonException, TwofactorException, QrTokenGenerationFailedException

_ = Translation('dataport-schuposh-twofactor-user-module').translate


def get_auth_token(url, username, password):
    r = requests.post(url + "/auth", json={'username': username, 'password': password})
    if r.status_code != 200:
        raise ValueError("wrong credentials")

    auth_token = r.json()["result"]["value"]["token"]
    if auth_token is None:
        raise ValueError("couldn't extract auth token")

    return auth_token


class Instance(Base):

    def init(self):
        try:
            config = ucr.ConfigRegistry()
            config.load()
            self.fulldebug = config.is_true("privacyidea/umc/fulldebug", False)
            self.url = config.get("privacyidea/saml/url")
            if self.url is None:
                raise ConfiguratonException

            # get privacyIDEA Credentials from /etc/twofactor-admin.secret
            config = configparser.RawConfigParser()
            config.read('/etc/twofactor-admin.secret')

            self.auth_token = get_auth_token(self.url, config.get('privacyIDEA', "user"), config.get('privacyIDEA', "pass"))
        except TwofactorException as e:
            raise UMC_Error(e.message)
        except configparser.NoSectionError:
            raise UMC_Error(_("PrivacyIDEA Credential Configuration Error, please contact Admin"))
        except ConnectionError:
            raise UMC_Error(_("Could not connect to the TokenService (PrivacyIDEA)"))

    def get_token_info(self):
        r = requests.get(self.url + "/token", headers={'Authorization': self.auth_token}, json={'user': self.username, 'active': "true"})
        if r.status_code == 400:
            raise UMC_Error("PrivacyIDEA cannot find the user %r" % (self.username,))
        elif r.status_code != 200:
            if r.json()['result']:
                MODULE.error("Error communicating with PrivacyIDEA:\n%r" % (r.json().result,))
            raise UMC_Error("Error communicating with PrivacyIDEA")
        tokens = r.json()['result']['value']['tokens']
        return tokens

    @simple_response
    def check_token_status(self):
        try:
            tokens = self.get_token_info()
            if len(tokens) > 0:
                return {"state": "user_has_token(s)"}
            else:
                return {"state": "user_has_no_token(s)"}
        except ConnectionError:
            raise UMC_Error("Connection to PrivacyIDEA failed")
        except TwofactorException as e:
            raise UMC_Error(e.message)

    @simple_response
    def generate_inital_qr_token(self):
        try:
            tokens = self.get_token_info()
            if len(tokens) != 0:
                # rare race condition where the token count check returned zero
                # and the user generated a token in another tab
                MODULE.error("rare race condition: len(tokens) == 0")
                raise QrTokenGenerationFailedException()

            MODULE.info('generate_inital_qr_token: got token')
            # generate qrcode
            r = requests.post(self.url + "/token/init", headers={'Authorization': self.auth_token},
                              json={'genkey': '1', 'user': self.username, 'type': 'totp', 'hashlib': 'sha1', "pin": ""})
            MODULE.info('generate_inital_qr_token: got response')
            result = r.json()
            try:
                img = result['detail']['googleurl']['img']
            except KeyError:
                img = None
            if img is None:
                MODULE.error("failed to call %s/token/init for %r" % (self.url, self.username, ))
                if self.fulldebug:
                    MODULE.error("failed to call %s/token/init for %r:\n%r" % (self.url, self.username, result,))
                raise QrTokenGenerationFailedException()
            MODULE.info('generate_inital_qr_token: got img')

            result = {"type": "initial_token", 'image': img}
            return result
        except ConnectionError:
            raise UMC_Error("Connection to PrivacyIDEA failed")
        except TwofactorException as e:
            raise UMC_Error(e.message)
