#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# dataport-schuposh-twofactor-user-module
# listener module
#
# Copyright 2021-2022 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.
#
# README:
#
# This module watches changes of the groups "lehrer-$OU" and "admins-$OU"
# and creates a group "mfa-user-$OU" with combined group members.
# If the group "mfa-user-$OU" is missing, it is created automatically.
# All group changes are collected and after 15 seconds of "silence", the
# required group modifications are performed only once per OU.

from __future__ import absolute_import

import listener
import univention.debug
import univention.admin.uldap
from ldap.filter import filter_format
from typing import Optional


name = "dataport-schuposh-twofactor-user-module"
description = "Manage SchoolSH two-factor user groups per school"
filter = "(&(|(ucsschoolRole=school_admin_group:school:*)(ucsschoolRole=school_teacher_group:school:*))(objectClass=univentionGroup))"


ldap_base = listener.configRegistry["ldap/base"]
domain_name = listener.configRegistry["domainname"]
server_role = listener.configRegistry["server/role"]


GRP_PREFIX = "mfa-user-"

TEACHER_PREFIX = (
    listener.configRegistry["ucsschool/ldap/default/groupprefix/teachers"]
).encode()
if not TEACHER_PREFIX:
    TEACHER_PREFIX = b"lehrer-"

ADMINS_PREFIX = (
    listener.configRegistry["ucsschool/ldap/default/groupprefix/admins"]
).encode()
if not ADMINS_PREFIX:
    ADMINS_PREFIX = b"admins-"

pending_ous = set()


def handler(dn, new, old):
    global pending_ous

    if server_role != "domaincontroller_master":
        return

    # get common name of old or new object
    cn = new.get("cn", [None])[0] or old.get("cn", [None])[0]
    if not cn:
        # should not happen :-)
        univention.debug.debug(
            univention.debug.LISTENER,
            univention.debug.ERROR,
            "{}: cn is not set: {}".format(name, dn),
        )
        return

    # if group prefix matches, add item to list of pending OUs
    if cn.startswith(TEACHER_PREFIX):
        pending_ous.add(cn[len(TEACHER_PREFIX):])
    if cn.startswith(ADMINS_PREFIX):
        pending_ous.add(cn[len(ADMINS_PREFIX):])


def get_group_members(
    lo, ou
):  # type: (univention.admin.uldap.Access, str) -> Optional[set[bytes]]
    univention.debug.debug(
        univention.debug.LISTENER,
        univention.debug.INFO,
        "{}: get_group_members({!r})".format(name, ou),
    )
    filter_str = filter_format(
        "(&(cn=%s%s)(objectClass=univentionGroup))",
        (GRP_PREFIX, ou)
    )
    result = lo.search(filter=filter_str, required=False)
    univention.debug.debug(
        univention.debug.LISTENER,
        univention.debug.INFO,
        "{}: get_group_members={!r}".format(name, result),
    )
    if not result:
        # ==> no object found
        return None
    return set(result[0][1].get("uniqueMember", []))


def create_grp(
    lo, grpmodule, ou
):  # type: (univention.admin.uldap.Access, any, str) -> None
    univention.debug.debug(
        univention.debug.LISTENER,
        univention.debug.PROCESS,
        "{}: create group for OU {!r}".format(name, ou),
    )
    grp_position = univention.admin.uldap.position(ldap_base)
    grp_position.setDn(
        "cn=groups,ou=%s,%s"
        % (
            ou,
            ldap_base,
        )
    )
    grpobj = grpmodule.object({}, lo, grp_position)
    grpobj.open()
    grpobj["name"] = "%s%s" % (GRP_PREFIX, ou.lower())
    grpobj["description"] = "contains users enabled for 2FA user UMC module"
    grpobj.create()


def update_grp(
    lo, grpmodule, grp_dn, members
):  # type: (univention.admin.uldap.Access, any, str, list[str]) -> None
    univention.debug.debug(
        univention.debug.LISTENER,
        univention.debug.PROCESS,
        "{}: update group {!r}".format(name, grp_dn),
    )
    grpobj = univention.admin.objects.get(grpmodule, {}, lo, position="", dn=grp_dn)
    grpobj.open()
    univention.debug.debug(
        univention.debug.LISTENER,
        univention.debug.INFO,
        "{}: update group: members={!r}".format(name, members),
    )
    grpobj["users"] = members
    grpobj.modify()


def postrun():  # type: () -> None
    global pending_ous

    listener.setuid(0)
    try:
        lo, po = univention.admin.uldap.getAdminConnection()
        univention.admin.modules.update()
        groupmod = univention.admin.modules.get("groups/group")
        univention.admin.modules.init(lo, po, groupmod)
    finally:
        listener.unsetuid()

    for ou in pending_ous:
        group_dn = "cn=%s%s,cn=groups,ou=%s,%s" % (
            GRP_PREFIX,
            ou.lower(),
            ou,
            ldap_base,
        )
        # bytes → str
        ou = ou.decode()

        grp_members = get_group_members(lo, ou.lower())  # type: set[bytes]

        if grp_members is None:
            # group has not been found ==> create one
            try:
                create_grp(lo, groupmod, ou)
            except Exception:
                import traceback

                univention.debug.debug(
                    univention.debug.ADMIN,
                    univention.debug.ERROR,
                    "CANNOT CREATE GROUP FOR OU %r!\n%s" % (ou, traceback.format_exc()),
                )
                continue
            grp_members = set()

        # get group members and merge them
        teacher_dn = "cn=%s%s,cn=groups,ou=%s,%s" % (TEACHER_PREFIX, ou, ou, ldap_base)
        admins_dn = "cn=%s%s,cn=ouadmins,cn=groups,%s" % (ADMINS_PREFIX, ou, ldap_base)
        teacher_members = set(
            lo.get(teacher_dn, ["uniqueMember"], required=False).get("uniqueMember", [])
        )  # type: set[bytes]
        admins_members = set(
            lo.get(admins_dn, ["uniqueMember"], required=False).get("uniqueMember", [])
        )  # type: set[bytes]
        joined_members = teacher_members | admins_members

        # members are still the same? then jump to next pending OU
        if joined_members == grp_members:
            continue

        # convert Set[bytes] to List[str]
        joined_members_str = [str(i) for i in list(joined_members)]
        # write new member list to group
        update_grp(lo, groupmod, group_dn, joined_members_str)

    # reset pending OU list
    pending_ous.clear()

# vim: set et ts=4 sw=4 :
